package envoy.client.event;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * This class handles events by allowing {@link EventHandler} object to register
 * themselves and then be notified about certain events dispatched by the event
 * bus.<br>
 * <br>
 * The event bus is a singleton and can be used across the entire application to
 * guarantee the propagation of events.
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>EventBus.java</strong><br>
 * Created: <strong>04.12.2019</strong><br>
 * 
 * @author Kai S. K. Engelbart
 * @since Envoy v0.2-alpha
 */
public class EventBus {

	/**
	 * Contains all {@link EventHandler} instances registered at this
	 * {@link EventBus}.
	 */
	private List<EventHandler> handlers = new ArrayList<>();

	/**
	 * The singleton instance of this {@link EventBus} that is used across the
	 * entire application.
	 */
	private static EventBus eventBus = new EventBus();

	/**
	 * This constructor is not accessible from outside this class because a
	 * singleton instance of it is provided by the {@link EventBus#getInstance()}
	 * method.
	 */
	private EventBus() {}

	/**
	 * @return the singleton instance of the {@link EventBus}
	 * @since Envoy v0.2-alpha
	 */
	public static EventBus getInstance() { return eventBus; }

	/**
	 * Registers a list of {@link EventHandler} objects to be notified when a
	 * {@link Event} is dispatched that they are subscribed to.
	 * 
	 * @param handlers the {@link EventHandler} objects to register
	 * @since Envoy v0.2-alpha
	 */
	public void register(EventHandler... handlers) { this.handlers.addAll(Arrays.asList(handlers)); }

	/**
	 * Dispatches a {@link Event} to every {@link EventHandler} subscribed to it.
	 * 
	 * @param event the {@link Event} to dispatch
	 * @since Envoy v0.2-alpha
	 */
	public void dispatch(Event<?> event) { handlers.stream().filter(h -> h.supports().contains(event.getClass())).forEach(h -> h.handle(event)); }

	/**
	 * @return a list of all {@link EventHandler} instances currently registered at
	 *         this {@link EventBus}
	 * @since Envoy v0.2-alpha
	 */
	public List<EventHandler> getHandlers() { return handlers; }
}
